#!/usr/bin/python
#-*- coding: utf-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.

 MQ² test script
"""

import os
import shutil
import sys
import unittest

from datetime import date

sys.path.insert(0, os.path.abspath('../'))

import MQ2
import MQ2.mq2 as mq2
from MQ2.add_marker_to_qtls import add_marker_to_qtls
from MQ2.add_qtl_to_map import add_qtl_to_map


TEST_INPUT_PASSED = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),'mapqtl', 'Demoset1.zip')
TEST_INPUT_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'csv', 'rqtl_out.csv')
TEST_INPUT_FAKE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),'invalid', 'fake.zip')

TEST_INPUT_FAILED = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),'invalid', 'multi_plugin.zip')

TEST_FOLDER = os.path.dirname(os.path.abspath(__file__))


def clean_directory():
    """ Remove the 'demo_test' folder is present in the current
    working directory.
    """
    if os.path.exists('demo_test') and os.path.isdir('demo_test'):
        shutil.rmtree('demo_test')
    for filename in os.listdir('.'):
        if filename.endswith(('.map', '.xls', '.xlsx', '.csv')):
            os.unlink(filename)


def read_file(filename):
    """ Reads in a file for a given filename and just return the
    content.
    """
    stream = open(filename)
    data = stream.read()
    stream.close()
    return data


class MQ2tests(unittest.TestCase):
    """ MQ² tests. """

    def __init__(self, method_name='runTest'):
        """ Constructor. """
        unittest.TestCase.__init__(self, method_name)
        self.mapqtl_folder = MQ2.set_tmp_folder()

    def setUp(self):
        """ Set up the environnment, ran before every tests. """
        clean_directory()
        if os.path.exists(self.mapqtl_folder):
            shutil.rmtree(self.mapqtl_folder)
        self.assertFalse(os.path.exists(self.mapqtl_folder))

    def tearDown(self):
        """ Clean up the environnment, ran after every tests. """
        clean_directory()
        if os.path.exists(self.mapqtl_folder):
            shutil.rmtree(self.mapqtl_folder)
        self.assertFalse(os.path.exists(self.mapqtl_folder))

    def test_extract_zip(self):
        """ Test the extract_zip function. """

        MQ2.extract_zip(TEST_INPUT_PASSED, self.mapqtl_folder)
        self.assertTrue(os.path.exists(self.mapqtl_folder))

        expected_files = ['Session 2 (IM)_A_trait01.mqo',
            'Session 2 (IM)_A_trait02.mqo']
        found_files = os.listdir(self.mapqtl_folder)
        found_files.sort()
        self.assertEqual(found_files, expected_files)

    def test_mq2_fake_zip(self):
        """ Test MQ² with a fake zip file """
        mapqtl_folder = MQ2.set_tmp_folder()
        MQ2.extract_zip(TEST_INPUT_FAKE, self.mapqtl_folder)
        self.assertTrue(os.path.exists(self.mapqtl_folder))

        found_files = os.listdir(self.mapqtl_folder)
        self.assertTrue(found_files == [])

    def test_mq2_get_matrix_dimensions(self):
        """ Test MQ² get_matrix_dimensions function """
        (length, width) = MQ2.get_matrix_dimensions(
            os.path.join(TEST_FOLDER,'mapqtl', 'qtls.exp'))
        self.assertEqual(length, 4)
        self.assertEqual(width, 14)

    def test_mq2_read_input_file(self):
        """ Test MQ² read_input_file function """
        matrix = MQ2.read_input_file(
            os.path.join(TEST_FOLDER, 'mapqtl', 'qtls.exp'), sep=',')
        self.assertEqual(len(matrix), 4)
        self.assertEqual(len(matrix[0]), 14)

    def test_MQ2_write_matrix(self):
        """ Test MQ² write_matrix function """
        data = [[1,2,3], [1,2,3,4], 'test']
        datafile = os.path.join(TEST_FOLDER, 'test_matrix.csv')

        self.assertFalse(os.path.exists(datafile))

        MQ2.write_matrix(datafile, data)

        self.assertTrue(os.path.exists(datafile))
        matrix = MQ2.read_input_file(datafile, sep=',')
        self.assertEqual(len(matrix), 3)
        self.assertEqual(len(matrix[0]), 3)
        self.assertEqual(len(matrix[1]), 4)
        self.assertEqual(len(matrix[2]), 1)

        os.unlink(datafile)

        self.assertFalse(os.path.exists(datafile))

    def test_get_plugin_and_folder_too_many_inputs(self):
        """ Test the get_plugin_and_folder function with too many inputs
        """
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.get_plugin_and_folder,
            inputzip=TEST_INPUT_PASSED,
            inputfile=TEST_INPUT_FILE)

    def test_get_plugin_and_folder_no_input(self):
        """ Test the get_plugin_and_folder function with no input
        """
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.get_plugin_and_folder)

    def test_get_plugin_and_folder_with_folder(self):
        """ Test the get_plugin_and_folder function with too many inputs
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)

        plugin2, folder2 = mq2.get_plugin_and_folder(
            inputdir=folder)
        self.assertEqual(plugin, plugin2)
        self.assertEqual(folder, folder2)

    def test_get_plugin_and_folder_multiple_files_supported(self):
        """ Test the get_plugin_and_folder function with a dataset
        containing different supported files.
        """
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.get_plugin_and_folder,
            inputzip=TEST_INPUT_FAILED)


if __name__ == '__main__':
    SUITE = unittest.TestLoader().loadTestsFromTestCase(MQ2tests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2 CSV plugin
"""

import os

from MQ2 import (MQ2Exception, MQ2NoSessionException,
                 MQ2NoSuchSessionException, MQ2NoMatrixException,
                 read_input_file, write_matrix)
from MQ2.plugin_interface import PluginInterface


def get_qtls_matrix(qtl_matrix, matrix, inputfile):
    """Extract for each position the LOD value obtained and save it in a
    matrix.
    This assumes that the first 4 columns are identical accross all mqo
    files (ie: the Group, Position and Locus are the same). This
    assumption should hold true if the files were generated from the
    same map.

    :arg qtl_matrix, the matrix in which to save the output.
    :arg matrix, the MapQTL file read in memory.
    :arg inputfile, name of the inputfile in which the QTLs have been
        found.

    """
    trait_name = inputfile.split(')_', 1)[1].split('.mqo')[0]
    matrix = list(zip(*matrix))
    if matrix[4][0] != 'LOD':
        raise MQ2Exception(
                'The file "%s" is not supported by MQ2. It may contain an '
                'analysis which does not return LOD values '
                '(such as Kruskal-Wallis or permutation test).' % inputfile)

    if not qtl_matrix:
        qtl_matrix = matrix[:4]
    else:
        if matrix[:4] != qtl_matrix[:4]:
            raise MQ2NoMatrixException(
                'The map used in the file "%s" does not'
                ' correspond to the map used in at least one other file.'
                % inputfile)
    tmp = list(matrix[4])
    tmp[0] = trait_name
    qtl_matrix.append(tmp)
    return qtl_matrix


def get_map_matrix(inputfile):
    """ Return the matrix representation of the genetic map.

    :arg inputfile: the path to the input file from which to retrieve the
        genetic map.

    """
    matrix = read_input_file(inputfile)
    output = []
    for row in matrix:
        if row[3]:
            output.append([row[3], row[1], row[2]])
    return output


def get_qtls_from_mapqtl_data(matrix, threshold, inputfile):
    """Extract the QTLs found by MapQTL reading its file.
    This assume that there is only one QTL per linkage group.

    :arg matrix, the MapQTL file read in memory
    :arg threshold, threshold used to determine if a given LOD value is
        reflective the presence of a QTL.
    :arg inputfile, name of the inputfile in which the QTLs have been
        found

    """
    trait_name = inputfile.split(')_', 1)[1].split('.mqo')[0]
    qtls = []
    qtl = None
    for entry in matrix[1:]:
        if qtl is None:
            qtl = entry
        if qtl[1] != entry[1]:
            if float(qtl[4]) > float(threshold):
                qtl[0] = trait_name
                qtls.append(qtl)
            qtl = None
            continue
        if entry[4] == '':  # pragma: no cover
            entry[4] = 0
        if qtl[4] == '':  # pragma: no cover
            qtl[4] = 0
        if float(entry[4]) > float(qtl[4]):
            qtl = entry

    if float(qtl[4]) > float(threshold):
        qtl[0] = trait_name
        if qtl not in qtls:
            qtls.append(qtl)

    return qtls


class MapQTLPlugin(PluginInterface):
    """ Plugin to extract QTLs from MapQTL output files.

    """

    name = 'MapQTL plugin'
    session_name = 'MapQTL session'

    @classmethod
    def is_applicable(cls):
        """ Functions used to check whether the plugin can be used or
        not.
        The MapQTL plugin relies on stdlib only, nothing to check

        """
        return True

    @classmethod
    def valid_file(cls, filename):
        """ Check if the provided file is a valid file for this plugin.

        :arg filename: the path to the file to check.

        """
        return not os.path.isdir(filename) \
            and os.path.basename(filename).startswith('Session ') \
            and filename.endswith('.mqo')

    @classmethod
    def get_files(cls, folder, session_id=''):
        """ Retrieve the list of files the plugin can work on.
        Find this list based on the files name, files extension or even
        actually by reading in the file.
        If a session identifier is specified it will restrict the list
        of files returned to those with this session identifier in their
        name.

        :arg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.
        :kwarg session_id: the session identifier of the MapQTL output
            to process.

        """
        filelist = []
        if folder is None or not os.path.isdir(folder):
            return filelist
        if session_id is None:
            session_id = ''
        for root, dirs, files in os.walk(folder):
            for filename in files:
                if filename.startswith('Session %s' % session_id) \
                        and filename.endswith('.mqo'):
                    filename = os.path.join(root, filename)
                    filelist.append(filename)
        return filelist

    @classmethod
    def get_session_identifiers(cls, folder=None, inputfile=None):
        """ Retrieve the list of session identifiers contained in the
        data on the folder.

        :kwarg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.
        :kwarg inputfile: the path to the input file to use

        """
        sessions = []
        if folder is None or not os.path.isdir(folder):
            return sessions
        for root, dirs, files in os.walk(folder):
            for filename in files:
                if filename.startswith('Session ') \
                        and filename.endswith('.mqo'):
                    session = filename.split()[1]
                    if session not in sessions:
                        sessions.append(session)
        return sessions

    @classmethod
    def convert_inputfiles(cls,
                           folder=None,
                           inputfile=None,
                           session=None,
                           lod_threshold=None,
                           qtls_file='qtls.csv',
                           matrix_file='qtls_matrix.csv',
                           map_file='map.csv'):
        """ Convert the input files present in the given folder or
        inputfile.
        This method creates the matrix representation of the QTLs
        results providing for each marker position the LOD value found
        for each trait as well as a representation of the genetic map
        used in the experiment.
        The genetic map should be cleared of any markers added by the
        QTL mapping software.

        :kwarg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.
        :kwarg inputfile: the path to the input file to use
        :kwarg session: the session identifier used to identify which
            session to process
        :kwarg lod_threshold: the LOD threshold to apply to determine if
            a QTL is significant or not
        :kwarg qtls_file: a csv file containing the list of all the
            significant QTLs found in the analysis.
            The matrix is of type:
               trait, linkage group, position, Marker, LOD other columns
        :kwarg matrix_file: a csv file containing a matrix representation
            of the QTL data. This matrix is of type:
               marker, linkage group, position, trait1 lod, trait2, lod
        :kwarg map_file: a csv file containing the genetic map used
            in this experiment. The map is of structure:
               marker, linkage group, position

        """
        if folder is None and inputfile is None:
            raise MQ2Exception('You must specify either a folder or an '
                               'input file')

        sessions = cls.get_session_identifiers(folder)
        if session is None:
            raise MQ2NoSessionException(
                'The MapQTL plugin requires a session identifier to '
                'identify the session to process.'
                'Sessions are: %s' % ','.join(sessions))
        elif str(session) not in sessions:
            raise MQ2NoSuchSessionException(
                'The MapQTL session provided (%s) could not be found in the '
                'dataset. '
                'Sessions are: %s' % (session, ','.join(sessions)))

        if folder is not None:
            if not os.path.isdir(folder):  # pragma: no cover
                raise MQ2Exception('The specified folder is actually '
                                   'not a folder')
            else:
                inputfiles = cls.get_files(folder, session_id=session)

        if inputfile is not None:  # pragma: no cover
            if os.path.isdir(inputfile):
                raise MQ2Exception('The specified input file is actually '
                                   'a folder')
            else:
                inputfiles = [inputfile]

        try:
            lod_threshold = float(lod_threshold)
        except ValueError:
            raise MQ2Exception('LOD threshold should be a number')

        inputfiles.sort()

        # QTL matrix and QTL files
        qtl_matrix = []
        qtls = []
        filename = None
        for filename in inputfiles:
            matrix = read_input_file(filename)
            headers = matrix[0]
            qtl_matrix = get_qtls_matrix(qtl_matrix, matrix, filename)
            qtls.extend(get_qtls_from_mapqtl_data(matrix, lod_threshold,
                        filename))
        # format QTLs and write down the selection
        headers[0] = 'Trait name'
        qtls.insert(0, headers)
        write_matrix(qtls_file, qtls)

        # Write down the QTL matrix
        del(qtl_matrix[0])
        # Reorganize a couple of columns
        qtl_matrix.insert(0, qtl_matrix[2])
        del(qtl_matrix[3])
        # write output
        qtl_matrix = list(zip(*qtl_matrix))
        write_matrix(matrix_file, qtl_matrix)

        # Map matrix
        map_matrix = get_map_matrix(inputfiles[0])
        write_matrix(map_file, map_matrix)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2, a QTL object that can be used by the plugin and is used by the
    MapChart code.
"""


class QTL(object):
    """ This object represents the QTL information extracted from the
    qtl matrix and needed for the MapChart output.

    """

    def __init__(self):
        """ Default constructor for the QTL object. """
        self.trait = ''
        self.start_position = ''
        self.peak_start_position = ''
        self.peak_stop_position = ''
        self.stop_position = ''
        self.peak_lod = 0

    def to_string(self):
        """ Return the string as it should be presented in a MapChart
        input file.

        """
        return '%s   %s %s %s %s' % (
            self.trait, self.start_position, self.peak_start_position,
            self.peak_stop_position, self.stop_position)

    def __repr__(self):  # pragma: no cover
        """ String representation of the QTL object. """
        return 'QTL<trait: %s, start:%s, peak:%s - %s, stop:%s>' % (
            self.trait,
            self.start_position, self.peak_start_position,
            self.peak_stop_position, self.stop_position)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2, code to add the closest marker to a specified location
    (ie: QTL peak).
"""

import logging

from MQ2 import read_input_file, write_matrix

LOG = logging.getLogger('MQ2')


def add_marker_to_qtl(qtl, map_list):
    """Add the closest marker to the given QTL.

    :arg qtl: a row of the QTL list.
    :arg map_list: the genetic map containing the list of markers.

    """
    closest = ''
    diff = None
    for marker in map_list:
        if qtl[1] == marker[1]:
            tmp_diff = float(qtl[2]) - float(marker[2])
            if diff is None or abs(diff) > abs(tmp_diff):
                diff = tmp_diff
                closest = marker
    if closest != '':
        closest = closest[0]
    return closest


def add_marker_to_qtls(qtlfile, mapfile, outputfile='qtls_with_mk.csv'):
    """This function adds to a list of QTLs, the closest marker to the
    QTL peak.

    :arg qtlfile: a CSV list of all the QTLs found.
        The file should be structured as follow::
            Trait, Linkage group, position, other columns

        The other columns will not matter as long as the first three
        columns are as such.
    :arg mapfile: a CSV representation of the map used for the QTL
        mapping analysis.
        The file should be structured as follow::
            Marker, Linkage group, position
    :kwarg outputfile: the name of the output file in which the list of
        QTLs with their closest marker will be written.

    """
    qtl_list = read_input_file(qtlfile, ',')
    map_list = read_input_file(mapfile, ',')
    if not qtl_list or not map_list:  # pragma: no cover
        return
    qtl_list[0].append('Closest marker')
    qtls = []
    qtls.append(qtl_list[0])
    for qtl in qtl_list[1:]:
        qtl.append(add_marker_to_qtl(qtl, map_list))
        qtls.append(qtl)
    LOG.info('- %s QTLs processed in %s' % (len(qtls), qtlfile))
    write_matrix(outputfile, qtls)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2, code to add the number of significant QTLs found for each position
    of the genetic map.
"""

import logging

from MQ2 import read_input_file, write_matrix

LOG = logging.getLogger('MQ2')


def add_qtl_to_marker(marker, qtls):
    """Add the number of QTLs found for a given marker.

    :arg marker, the marker we are looking for the QTL's.
    :arg qtls, the list of all QTLs found.

    """
    cnt = 0
    for qtl in qtls:
        if qtl[-1] == marker[0]:
            cnt = cnt + 1

    marker.append(str(cnt))
    return marker


def add_qtl_to_map(qtlfile, mapfile, outputfile='map_with_qtls.csv'):
    """ This function adds to a genetic map for each marker the number
    of significant QTLs found.

    :arg qtlfile, the output from MapQTL transformed to a csv file via
        'parse_mapqtl_file' which contains the closest markers.
    :arg mapfile, the genetic map with all the markers.
    :kwarg outputfile, the name of the output file in which the map will
        be written.

    """
    qtl_list = read_input_file(qtlfile, ',')
    map_list = read_input_file(mapfile, ',')
    map_list[0].append('# QTLs')
    markers = []
    markers.append(map_list[0])
    qtl_cnt = 0
    for marker in map_list[1:]:
        markers.append(add_qtl_to_marker(marker, qtl_list[1:]))
        qtl_cnt = qtl_cnt + int(markers[-1][-1])
    LOG.info('- %s markers processed in %s' % (len(markers), mapfile))
    LOG.info('- %s QTLs located in the map: %s' % (qtl_cnt, outputfile))
    write_matrix(outputfile, markers)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2 Excel plugin
"""

import os
import re

# Check if import works
_VALID = False
SUPPORTED_FILES = []
try:  # pragma: no cover
    import xlrd
    _VALID = True
    SUPPORTED_FILES = ['xls']
    if xlrd.__VERSION__ >= '0.8.0':
        SUPPORTED_FILES.append('xlsx')
except ImportError:  # pragma: no cover
    pass


from MQ2 import (MQ2Exception, MQ2NoSessionException,
                 MQ2NoSuchSessionException, write_matrix)
from MQ2.plugin_interface import PluginInterface


def is_excel_file(inputfile):
    """ Return whether the provided file is a CSV file or not.
    This checks if the first row of the file can be splitted by ',' and
    if the resulting line contains more than 4 columns (Markers, linkage
    group, chromosome, trait).

    """
    try:
        xlrd.open_workbook(inputfile)
    except Exception as err:
        print(err)
        return False
    return True


def read_excel_file(inputfile, sheet_name):
    """ Return a matrix containing all the information present in the
    excel sheet of the specified excel document.

    :arg inputfile: excel document to read
    :arg sheetname: the name of the excel sheet to return

    """
    workbook = xlrd.open_workbook(inputfile)
    output = []
    found = False
    for sheet in workbook.sheets():
        if sheet.name == sheet_name:
            found = True
            for row in range(sheet.nrows):
                values = []
                for col in range(sheet.ncols):
                    values.append(sheet.cell(row, col).value)
                output.append(values)
    if not found:  # pragma: no cover
        raise MQ2Exception('Invalid session identifier provided')
    return output


def get_qtls_from_rqtl_data(matrix, lod_threshold):
    """ Retrieve the list of significants QTLs for the given input
    matrix and using the specified LOD threshold.
    This assumes one QTL per linkage group.

    :arg matrix, the MapQTL file read in memory
    :arg threshold, threshold used to determine if a given LOD value is
        reflective the presence of a QTL.

    """
    t_matrix = list(zip(*matrix))
    qtls = [['Trait', 'Linkage Group', 'Position', 'LOD', 'Exact marker']]
    # row 0: markers
    # row 1: chr
    # row 2: pos
    for row in t_matrix[3:]:
        lgroup = None
        max_lod = None
        peak = None
        cnt = 1
        while cnt < len(row):
            if lgroup is None:
                lgroup = t_matrix[1][cnt]

            if lgroup == t_matrix[1][cnt]:
                if max_lod is None:
                    max_lod = float(row[cnt])
                if float(row[cnt]) > float(max_lod):
                    max_lod = float(row[cnt])
                    peak = cnt
            else:
                if max_lod \
                        and float(max_lod) > float(lod_threshold) \
                        and peak:
                    qtl = [row[0],             # trait
                           t_matrix[1][peak],  # LG
                           t_matrix[2][peak],  # pos
                           max_lod,            # LOD value
                           t_matrix[0][peak],  # marker
                           ]
                    qtls.append(qtl)
                lgroup = None
                max_lod = None
                peak = cnt
            cnt = cnt + 1
    return qtls


def get_map_matrix(inputfile, sheet_name):
    """ Return the matrix representation of the genetic map.

    :arg inputfile: the path to the input file from which to retrieve the
        genetic map.
    :arg sheet_name: the excel sheet containing the data on which to
        retrieve the genetic map.

    """
    matrix = read_excel_file(inputfile, sheet_name)
    output = [['Locus', 'Group', 'Position']]
    for row in matrix:
        if row[0] and not re.match(r'c\d+\.loc[\d\.]+', row[0]):
            output.append([row[0], row[1], row[2]])
    return output


class XslPlugin(PluginInterface):
    """ Plugin to extract QTLs from a matrix CSV file.

    """

    name = 'Excel plugin'
    session_name = 'Excel sheet'

    @classmethod
    def is_applicable(cls):
        """ Functions used to check whether the plugin can be used or
        not.
        The CSV plugin relies on stdlib only, nothing to check

        """
        global _VALID
        return _VALID

    @classmethod
    def valid_file(cls, filename):
        """ Check if the provided file is a valid file for this plugin.

        :arg filename: the path to the file to check.

        """
        file_ex = os.path.splitext(filename)[1].replace('.', '', 1)
        return file_ex in SUPPORTED_FILES and is_excel_file(filename)

    @classmethod
    def get_files(cls, folder):
        """ Retrieve the list of files the plugin can work on.
        Find this list based on the files name, files extension or even
        actually by reading in the file.

        :arg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.

        """
        filelist = []
        if folder is None or not os.path.isdir(folder):
            return filelist
        for root, dirs, files in os.walk(folder):
            for filename in files:
                for ext in SUPPORTED_FILES:
                    filename = os.path.join(root, filename)
                    if filename.endswith(ext) and is_excel_file(filename):
                        filelist.append(filename)
        return filelist

    @classmethod
    def get_session_identifiers(cls, folder=None, inputfile=None):
        """ Retrieve the list of session identifiers contained in the
        data on the folder or the inputfile.
        For this plugin, it returns the list of excel sheet available.

        :kwarg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.
        :kwarg inputfile: the path to the input file to use

        """
        sessions = []
        if inputfile and folder:
            raise MQ2Exception(
                'You should specify either a folder or a file')
        if folder:
            if not os.path.isdir(folder):
                return sessions
            for root, dirs, files in os.walk(folder):
                for filename in files:
                    filename = os.path.join(root, filename)
                    for ext in SUPPORTED_FILES:
                        if filename.endswith(ext):
                            wbook = xlrd.open_workbook(filename)
                            for sheet in wbook.sheets():
                                if sheet.name not in sessions:
                                    sessions.append(sheet.name)
        elif inputfile:
            if os.path.isdir(inputfile):
                return sessions
            for ext in SUPPORTED_FILES:
                if inputfile.endswith(ext):
                    wbook = xlrd.open_workbook(inputfile)
                    for sheet in wbook.sheets():
                        if sheet.name not in sessions:
                            sessions.append(sheet.name)
        return sessions

    @classmethod
    def convert_inputfiles(cls,
                           folder=None,
                           inputfile=None,
                           session=None,
                           lod_threshold=None,
                           qtls_file='qtls.csv',
                           matrix_file='qtls_matrix.csv',
                           map_file='map.csv'):
        """ Convert the input files present in the given folder or
        inputfile.
        This method creates the matrix representation of the QTLs
        results providing for each marker position the LOD value found
        for each trait as well as a representation of the genetic map
        used in the experiment.
        The genetic map should be cleared of any markers added by the
        QTL mapping software.

        :kwarg folder: the path to the folder containing the files to
            check. This folder may contain sub-folders.
        :kwarg inputfile: the path to the input file to use
        :kwarg session: the session identifier used to identify which
            session to process
        :kwarg lod_threshold: the LOD threshold to apply to determine if
            a QTL is significant or not
        :kwarg qtls_file: a csv file containing the list of all the
            significant QTLs found in the analysis.
            The matrix is of type:
               trait, linkage group, position, Marker, LOD other columns
        :kwarg matrix_file: a csv file containing a matrix representation
            of the QTL data. This matrix is of type:
               marker, linkage group, position, trait1 lod, trait2, lod
        :kwarg map_file: a csv file containing the genetic map used
            in this experiment. The map is of structure:
               marker, linkage group, position

        """
        if folder is None and inputfile is None:
            raise MQ2Exception('You must specify either a folder or an '
                               'input file')

        if folder is not None:  # pragma: no cover
            if not os.path.isdir(folder):
                raise MQ2Exception('The specified folder is actually '
                                   'not a folder')
            else:
                inputfiles = cls.get_files(folder)

        if inputfile is not None:  # pragma: no cover
            if os.path.isdir(inputfile):
                raise MQ2Exception('The specified input file is actually '
                                   'a folder')
            else:
                inputfiles = [inputfile]

        sessions = cls.get_session_identifiers(
            folder=folder, inputfile=inputfile)

        if session is None:
            raise MQ2NoSessionException(
                'The Excel plugin requires a sheet identifier to '
                'identify the sheet of the workbook to process. '
                'Sheets are: %s' % ','.join(sessions))
        elif str(session) not in sessions:
            raise MQ2NoSuchSessionException(
                'The Excel sheet provided (%s) could not be found in the '
                'workbook. '
                'Sheets are: %s' % (session, ','.join(sessions)))

        if len(inputfiles) > 1:  # pragma: no cover
            raise MQ2Exception(
                'This plugin can only process one file at a time')

        try:
            lod_threshold = float(lod_threshold)
        except ValueError:
            raise MQ2Exception('LOD threshold should be a number')

        inputfile = inputfiles[0]

        # QTL matrix and QTL files
        qtls = []
        matrix = read_excel_file(inputfile, sheet_name=session)
        qtls.extend(get_qtls_from_rqtl_data(matrix, lod_threshold))
        # format QTLs and write down the selection
        write_matrix(qtls_file, qtls)

        # Write down the QTL matrix
        write_matrix(matrix_file, matrix)

        # Map matrix
        map_matrix = get_map_matrix(inputfile, session)
        write_matrix(map_file, map_matrix)



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2 main program
"""


import argparse
import logging
import os
import shutil
import sys

from straight.plugin import load

sys.path.insert(0, os.path.join(os.path.dirname(
                os.path.abspath(__file__)), '..'))

from MQ2 import (__version__,
                 set_tmp_folder,
                 extract_zip,
                 MQ2Exception,
                 read_input_file,
                 write_matrix)
from MQ2.plugin_interface import PluginInterface
from MQ2.add_marker_to_qtls import add_marker_to_qtls
from MQ2.add_qtl_to_map import add_qtl_to_map
from MQ2.mapchart import generate_map_chart_file


logging.basicConfig()
LOG = logging.getLogger('MQ2')


def _get_arguments():  # pragma: no cover
    """ Handle the command line arguments given to this program """
    LOG.debug('Parse command line argument')
    parser = argparse.ArgumentParser(
        description='Command line interface for the MQ² program')

    parser.add_argument(
        '-z', '--zipfile', dest='inputzip', default=None,
        help='Zip file containing the input files.')
    parser.add_argument(
        '-d', '--dir', dest='inputdir', default=None,
        help='Path to a local folder containing the input files.')
    parser.add_argument(
        '-f', '--file', dest='inputfile', default=None,
        help='Path to a local input file.')

    parser.add_argument(
        '--lod', default=3,
        help='LOD threshold to use to assess the significance of a LOD \
        value for a QTL.')

    parser.add_argument(
        '--session', default=None,
        help='Session to analyze if required.')

    parser.add_argument(
        '--verbose', action='store_true',
        help="Gives more info about what's going on")
    parser.add_argument(
        '--debug', action='store_true',
        help="Outputs debugging information")
    parser.add_argument(
        '--version', action='version',
        version='MQ² version: %s' % __version__)
    return parser.parse_args()


def cli_main():  # pragma: no cover
    """ Main function when running from CLI. """
    if '--debug' in sys.argv:
        LOG.setLevel(logging.DEBUG)
    elif '--verbose' in sys.argv:
        LOG.setLevel(logging.INFO)

    args = _get_arguments()
    try:
        plugin, folder = get_plugin_and_folder(
            inputzip=args.inputzip,
            inputdir=args.inputdir,
            inputfile=args.inputfile)
        LOG.debug('Plugin: %s -- Folder: %s' % (plugin.name, folder))
        run_mq2(
            plugin, folder, lod_threshold=args.lod, session=args.session)
    except MQ2Exception as err:
        print(err)
        return 1
    return 0


def get_plugin_and_folder(inputzip=None, inputdir=None, inputfile=None):
    """ Main function. """

    if (inputzip and inputdir) \
            or (inputzip and inputfile) \
            or (inputdir and inputfile):
        raise MQ2Exception('You must provide either a zip file or a '
                           'directory or an input file as input.')
    if not inputzip and not inputdir and not inputfile:
        raise MQ2Exception('You must provide either a zip file or a '
                           'directory or an input file as input.')

    # retrieve input: file, directory, zip
    if inputzip:
        tmp_folder = set_tmp_folder()
        extract_zip(inputzip, tmp_folder)
    elif inputfile:
        tmp_folder = inputfile
    else:
        tmp_folder = inputdir

    # retrieve the plugins
    plugins = load('MQ2.plugins', subclasses=PluginInterface)
    LOG.debug('Plugin loaded: %s' % [plugin.name for plugin in plugins])

    # keep only the plugins that will work
    plugins = [plugin for plugin in plugins if plugin.is_applicable()]
    LOG.debug('Plugin applicable: %s' % [plugin.name for plugin in plugins])

    # keep only the plugins that have the file(s) they need
    if inputfile:
        plugins = [plugin for plugin in plugins
                   if plugin.valid_file(tmp_folder)]
    else:
        plugins = [plugin for plugin in plugins
                   if plugin.get_files(tmp_folder)]

    LOG.debug('Plugin w/ valid input: %s' %
              [plugin.name for plugin in plugins])

    if len(plugins) > 1:
        raise MQ2Exception('Your dataset contains valid input for '
                           'several plugins.')
    if len(plugins) == 0:
        raise MQ2Exception('Invalid dataset: your input cannot not be '
                           'processed by any of the current plugins.')
    plugin = plugins[0]
    return (plugin, tmp_folder)


def run_mq2(plugin, folder, lod_threshold=None, session=None,
            outputfolder=None):
    """ Run the plugin. """

    qtls_file = 'qtls.csv'
    matrix_file = 'qtls_matrix.csv'
    map_file = 'map.csv'
    map_qtl_file = 'map_with_qtls.csv'
    qtls_ml_file = 'qtls_with_mk.csv'
    map_chart_file = 'MapChart.map'
    if outputfolder:  # pragma: no cover
        qtls_file = '%s/%s' % (outputfolder, qtls_file)
        qtls_ml_file = '%s/%s' % (outputfolder, qtls_ml_file)
        matrix_file = '%s/%s' % (outputfolder, matrix_file)
        map_file = '%s/%s' % (outputfolder, map_file)
        map_qtl_file = '%s/%s' % (outputfolder, map_qtl_file)
        map_chart_file = '%s/%s' % (outputfolder, map_chart_file)

    LOG.debug('Call the plugin to create the map, qtls and matrix files')
    if folder and os.path.isdir(folder):
        plugin.convert_inputfiles(folder=folder, session=session,
                                  lod_threshold=lod_threshold,
                                  qtls_file=qtls_file,
                                  matrix_file=matrix_file,
                                  map_file=map_file)
    else:
        plugin.convert_inputfiles(inputfile=folder, session=session,
                                  lod_threshold=lod_threshold,
                                  qtls_file=qtls_file,
                                  matrix_file=matrix_file,
                                  map_file=map_file)

    LOG.debug('Add the number of QTLs found on the matrix')
    _append_count_to_matrix(matrix_file, lod_threshold)

    LOG.debug('Append the closest marker to the peak')
    add_marker_to_qtls(qtls_file, map_file, outputfile=qtls_ml_file)

    LOG.debug('Put the number of QTLs found on each marker of the map')
    add_qtl_to_map(qtls_ml_file, map_file, outputfile=map_qtl_file)

    LOG.debug('Generate the mapchart file')
    generate_map_chart_file(matrix_file, lod_threshold,
                            map_chart_file=map_chart_file)

    if folder and os.path.isdir(folder) and os.path.exists(folder):
        shutil.rmtree(folder)
    return 0


def _append_count_to_matrix(qtl_matrixfile, lod_threshold):
    """ Append an extra column at the end of the matrix file containing
    for each row (marker) the number of QTL found if the marker is known
    ie: Locus != ''

    :arg qtl_matrix, the matrix in which to save the output.
    :arg threshold, threshold used to determine if a given LOD value is
        reflective the presence of a QTL.

    """
    if not os.path.exists(qtl_matrixfile):  # pragma: no cover
        raise MQ2Exception('File not found: "%s"' % qtl_matrixfile)
    matrix = read_input_file(qtl_matrixfile, sep=',')
    tmp = list(matrix[0])
    tmp.append('# QTLs')
    matrix[0] = tmp
    cnt = 1
    while cnt < len(matrix):
        row = list(matrix[cnt])
        nr_qtl = 0
        for cel in row[3:]:
            if cel and float(cel) > float(lod_threshold):
                nr_qtl = nr_qtl + 1
        row.append(str(nr_qtl))
        matrix[cnt] = row
        cnt = cnt + 1
    write_matrix(qtl_matrixfile, matrix)


if __name__ == "__main__":  # pragma: no cover
    cli_main()



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2 plugin interface
"""


class PluginInterface(object):  # pragma: no cover
    """ The interface that each plugin should extends to support their
    file format and tool.

    Each plugin should be able to detect if it can be run or not
    automatically. In case the plugin cannot be used, it should not
    prevent the other plugins from running.

    Each plugin should be able to detect in a folder (which may contain
    subfolders) if there are files they are compatible with.

    For some plugin, multiple analyzes can be submitted at once and the
    plugin needs to know which one should be analyzed.
    This will be the role of the ``session`` argument.

    """

    name = 'plugin name'
    session_name = 'The name of the session to be displayed on the form'

    @classmethod
    def is_applicable(cls):
        """ Functions used to check whether the plugin can be used or
        not.
        This is the function that would check the import and that should
        make sure the rest of the plugin will run smoothly.

        The method ``is_applicable`` will be called by the program and
        should return a simple boolean telling if the plugin can be run
        or not.
        It is this method that should check that all the potential
        dependencies are met for the plugin to run.

        """
        pass

    @classmethod
    def get_files(cls, folder):
        """ Retrieve the list of files the plugin can work on.
        Find this list based on the files name, files extension or even
        actually by reading in the file.

        The method ``get_files`` will browse the provided path for any
        file in the specified folder or sub-folder that the plugin can
        handle.

        .. note:: ``get_files`` should be able to handle the case where
                  the provided path points to a file rather than a
                  folder, in which case the plugin should return an
                  empty list.

        :arg folder: string of the path to the folder containing the
            files to check. This folder may contain sub-folders.

        """
        pass

    @classmethod
    def valid_file(cls, filename):
        """ Check if the provided file is a valid file for this plugin.

        Since MQ² can also be used on a single file via the command-line,
        the ``valid_file`` will then be used to check if the provided
        file can be handled by the plugin.

        .. note:: ``valid_file`` should be able to handle the case where
                  the provided path points to a director rather than a
                  file, in which case the plugin should return a `False`
                  boolean.

        :arg filename: string of the path to the file to check.

        """
        pass

    @classmethod
    def get_session_identifiers(cls, folder=None, inputfile=None):
        """ Retrieve the list of session identifiers contained in the
        data on the folder or the inputfile.

        The method ``get_session_identifiers`` is used by the
        web-interface to present to the user a list of sessions they can
        choose from and by the command line interface, when the user did
        not specified a session or specified an invalid session.

        The ``get_session_identifiers`` method receive either a folder
        or an inputfile argument (and should raise and
        :class:`~MQ2.MQ2Exception` if both are provided). The method
        extract the session identifiers from this input and return them
        as a list. If both ``folder`` and ``inputfile`` are ``None``, it
        may return an empty list.

        :kwarg folder: string of the path to the folder containing the
            files to check. This folder may contain sub-folders.
        :kwarg inputfile: string of the path to the input file to use

        """
        pass

    @classmethod
    def convert_inputfiles(cls,
                           folder=None,
                           inputfile=None,
                           session=None,
                           lod_threshold=None,
                           qtls_file='qtls.csv',
                           matrix_file='qtls_matrix.csv',
                           map_file='map.csv'):
        """ Convert the input files present in the given folder or
        inputfile.
        This method creates the matrix representation of the QTLs
        results providing for each marker position the LOD value found
        for each trait as well as a list of all the significant QTLs
        found in the results and a representation of the genetic map
        used in the experiment.
        The genetic map should be cleared of any markers added by the
        QTL mapping software.

        :kwarg folder: string of the path to the folder containing the
            files to check. This folder may contain sub-folders.
        :kwarg inputfile: string of the path to the input file to use
        :kwarg session: the session identifier used to identify which
            session to process
        :kwarg lod_threshold: the LOD threshold to apply to determine if
            a QTL is significant or not
        :kwarg qtls_file: a csv file containing the list of all the
            significant QTLs found in the analysis.
            The matrix is of type:
            ``trait, linkage group, position, Marker, LOD other columns``
        :kwarg matrix_file: a csv file containing a matrix representation
            of the QTL data. This matrix is of type:
            ``marker, linkage group, position, trait1 lod, trait2, lod``
        :kwarg map_file: a csv file containing the genetic map used
            in this experiment. The map is of structure:
            ``marker, linkage group, position``

        """
        pass



#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""


import datetime
import logging
import os
import tarfile
import tempfile
import zipfile

LOG = logging.getLogger('MQ2')

__version__ = '1.0.0'


def set_tmp_folder():
    """ Create a temporary folder using the current time in which
    the zip can be extracted and which should be destroyed afterward.
    """
    output = "%s" % datetime.datetime.now()
    for char in [' ', ':', '.', '-']:
        output = output.replace(char, '')
    output.strip()
    tmp_folder = os.path.join(tempfile.gettempdir(), output)
    return tmp_folder


def extract_zip(filename, extract_dir):
    """ Extract the sources in a temporary folder.
    :arg filename, name of the zip file containing the data from MapQTL
    which will be extracted
    :arg extract_dir, folder in which to extract the archive.
    """
    LOG.info("Extracting %s in %s " % (filename, extract_dir))
    if not os.path.exists(extract_dir):
        try:
            os.mkdir(extract_dir)
        except IOError as err:  # pragma: no cover
            LOG.info("Could not generate the folder %s" % extract_dir)
            LOG.debug("Error: %s" % err)
            return

    if zipfile.is_zipfile(filename):
        try:
            zfile = zipfile.ZipFile(filename, "r")
            for name in zfile.namelist():
                if os.path.dirname(name):
                    curdir = os.path.join(extract_dir, os.path.dirname(name))
                    if not os.path.exists(curdir):
                        os.mkdir(curdir)
                        continue
                outfile = open(os.path.join(extract_dir, name), 'wb')
                outfile.write(zfile.read(name))
                outfile.flush()
                outfile.close()
            zfile.close()
        except IOError as err:  # pragma: no cover
            LOG.info("Error while extracting the zip archive.")
            LOG.debug("Error: %s" % err)
    else:
        try:  # pragma: no cover  We only have zipfile to test with
            tar = tarfile.open(filename)
            tar.extractall(extract_dir)
            tar.close()
        except tarfile.ReadError as err:  # pragma: no cover
            LOG.info("Error while extracting the tarball.")
            LOG.debug("Error: %s" % err)

    return extract_dir


def get_matrix_dimensions(filename):
    """ Reads in a matrix file (comma separated) and returns the number
    of rows and columns.
    :arg filename, the full path to the file to read.
    """
    stream = None
    try:
        stream = open(filename, 'r')
        length = len(stream.readlines())
        stream.seek(0)
        width = len(stream.readline().split(','))
        return (length, width)
    except IOError as err:  # pragma: no cover
        LOG.info("Something wrong happend while reading the file %s "
                 % filename)
        LOG.debug("ERROR: %s" % err)
    finally:
        if stream:
            stream.close()


def read_input_file(filename, sep='\t', noquote=False):
    """Reads a given inputfile (tab delimited) and returns a matrix
    (list of list).
    arg: filename, the complete path to the inputfile to read
    """
    output = []
    stream = None
    try:
        stream = open(filename, 'r')
        for row in stream:
            row = row.strip()
            if noquote:
                row = row.replace('"', '')
            output.append(row.split(sep))
    except IOError as err:  # pragma: no cover
        LOG.info("Something wrong happend while reading the file %s "
                 % filename)
        LOG.debug("ERROR: %s" % err)
    finally:
        if stream:
            stream.close()
    return output


def write_matrix(outputfile, matrix):
    """ Write down the provided matrix in the specified outputfile.
    :arg outputfile, name of the outputfile in which the QTLs found are
        written.
    :arg matrix, the list of lists of data to write.
    """

    try:
        stream = open(outputfile, 'w')
        for row in matrix:
            if isinstance(row, list) or isinstance(row, tuple):
                row = [str(el).strip() for el in row]
                stream.write(','.join(row) + '\n')
            else:
                stream.write(row + '\n')
    except IOError as err:  # pragma: no cover
        LOG.info('An error occured while writing the file %s'
                 % outputfile)
        LOG.debug("Error: %s" % err)
    finally:
        stream.close()
    LOG.info('Wrote QTLs in file %s' % outputfile)


class MQ2Exception(Exception):
    """ Basic exception class to be used by the pymq2 library. """
    pass


class MQ2NoMatrixException(MQ2Exception):
    """ Exception raised when the qtl_matrix file can not be generated
    because the maps used in the mqo files are inconsistent.
    """
    pass


class MQ2NoSessionException(MQ2Exception):
    """ Exception raised when there are no MapQTL files corresponding to
    the session identifier provided by the user.
    """
    pass


class MQ2NoSuchSessionException(MQ2Exception):
    """ Exception raised when there are no MapQTL files corresponding to
    the session identifier provided by the user.
    """
    pass



#!/usr/bin/python
#-*- coding: UTF-8 -*-

"""
Setup script
"""

from setuptools import setup
from MQ2 import __version__

setup(
    name='MQ2',
    description='MQ² aims at extracting and find QTLs from MapQTL output',
    author='Pierre-Yves Chibon',
    author_email='pingou@pingoured.fr',
    version=__version__,
    license='GPLv3+',
    url='https://github.com/PBR/MQ2/',
    packages=['MQ2', 'MQ2.plugins'],
    install_requires=['straight.plugin', 'xlrd'],
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'MQ2 = MQ2.mq2:cli_main'
		]
    },
    classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          ],
    )




#-*- coding: UTF-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.
"""

"""
MQ2 map chart generation from the data matrix.
"""

import logging

from MQ2 import read_input_file
from MQ2.qtl import QTL


LOG = logging.getLogger('MQ2')


def _extrac_qtl(peak, block):
    """ Given a row containing the peak of the QTL and all the rows of
    the linkage group of the said QTL (splitted per trait), determine
    the QTL interval and find the start and stop marker of the said
    interval.
    The interval is a LOD 2 interval.
    The approach is conservative in the way it takes the first and last
    marker within the interval.

    :arg peak, a list containing the row information for the peak marker
    :arg block, a list containing all the rows in the linkage group of
        this QTL, splitted per trait.

    """
    qtls = []
    if not peak:
        return qtls
    for trait in peak:
        threshold = 2
        # Search QTL start
        cnt = block.index(peak[trait])
        start = block[cnt]
        while cnt > 0:
            if block[cnt][-2] == trait:
                if (float(block[cnt][-1]) - float(threshold)) >= 0:
                    start = block[cnt]
            cnt = cnt - 1

        # Search QTL end
        end = []
        cnt = block.index(peak[trait])
        end = block[cnt]
        while cnt < len(block):
            if block[cnt][-2] == trait:
                if (float(block[cnt][-1]) - float(threshold)) >= 0:
                    end = block[cnt]
            cnt = cnt + 1

        qtl = QTL()
        qtl.trait = trait
        qtl.start_position = start[1]
        qtl.peak_start_position = peak[trait][1]
        qtl.peak_stop_position = peak[trait][1]
        qtl.stop_position = end[1]
        qtls.append(qtl)
    return qtls


def _order_linkage_group(group):
    """ For a given group (ie: a list containing [marker, position])
    order the list according to their position.
    """
    tmp = {}
    for row in group:
        if float(row[1]) in tmp:  # pragma: no cover
            tmp[float(row[1])].append(row[0])
        else:
            tmp[float(row[1])] = [row[0]]

    keys = list(tmp.keys())
    keys.sort()
    output = []
    for key in keys:
        for entry in tmp[key]:
            if not entry:
                continue
            output.append([entry, str(key)])
    return output


def generate_map_chart_file(qtl_matrix, lod_threshold,
                            map_chart_file='MapChart.map'):
    """ This function converts our QTL matrix file into a MapChart input
    file.

    :arg qtl_matrix: the path to the QTL matrix file generated by
        the plugin.
    :arg lod_threshold: threshold used to determine if a given LOD value
        is reflective the presence of a QTL.
    :kwarg map_chart_file: name of the output file containing the
        MapChart information.

    """

    qtl_matrix = read_input_file(qtl_matrix, sep=',')
    tmp_dic = {}
    cnt = 1
    tmp = {}
    block = []
    for row in qtl_matrix[1:]:
        linkgrp = qtl_matrix[cnt - 1][1]
        if cnt == 1:
            linkgrp = qtl_matrix[cnt][1]

        if not linkgrp in tmp_dic:
            tmp_dic[linkgrp] = [[], []]

        infos = row[1:4]
        if qtl_matrix[cnt][1] != linkgrp:
            qtls = _extrac_qtl(tmp, block)
            tmp_dic[linkgrp][1] = qtls
            linkgrp = qtl_matrix[cnt][1]
            tmp_dic[linkgrp] = [[], []]
            tmp = {}
            block = []

        tmp_dic[linkgrp][0].append([row[0], row[2]])

        colcnt = 4
        for cel in row[4:-1]:
            blockrow = infos[:]
            blockrow.extend([qtl_matrix[0][colcnt], cel])
            block.append(blockrow)
            if cel.strip() != '' and float(cel) >= float(lod_threshold):
                temp = infos[:]
                if not tmp:
                    temp.extend([qtl_matrix[0][colcnt], cel])
                    tmp[qtl_matrix[0][colcnt]] = temp
                elif (qtl_matrix[0][colcnt] in tmp
                      and float(cel) >= float(
                      tmp[qtl_matrix[0][colcnt]][-1])) \
                        or qtl_matrix[0][colcnt] not in tmp:
                    temp.extend([qtl_matrix[0][colcnt], cel])
                    tmp[qtl_matrix[0][colcnt]] = temp
            colcnt = colcnt + 1
        cnt = cnt + 1

    try:
        stream = open(map_chart_file, 'w')
        keys = list(tmp_dic.keys())
        ## Remove unknown group, reason:
        # The unlinked markers, if present, are always put in group U by
        # MapQTL. If you don't omit them and there are many (often), then
        # their names take so much space that it is difficult to fit them
        # on the page.
        if 'U' in keys:
            keys.remove('U')
        # Try to convert all the groups to float, which would result in
        # a better sorting. If that fails, fail silently.
        try:
            keys = [int(key) for key in keys]
        except ValueError:
            pass
        keys.sort()
        for key in keys:
            key = str(key)  # Needed since we might have converted them to int
            if tmp_dic[key]:
                if key == 'U':  # pragma: no cover
                    # We removed the key before, we should not be here
                    continue
                stream.write('group %s\n' % key)
                for entry in _order_linkage_group(tmp_dic[key][0]):
                    stream.write('  '.join(entry) + '\n')
                if tmp_dic[key][1]:
                    stream.write('\n')
                    stream.write('qtls\n')
                    for qtl in tmp_dic[key][1]:
                        stream.write('%s \n' % qtl.to_string())
                stream.write('\n')
                stream.write('\n')
    except IOError as err:  # pragma: no cover
        LOG.info('An error occured while writing the map chart map '
                 'to the file %s' % map_chart_file)
        LOG.debug("Error: %s" % err)
    finally:
        stream.close()
    LOG.info('Wrote MapChart map in file %s' % map_chart_file)



#!/usr/bin/python
#-*- coding: utf-8 -*-

"""
 (c) 2011-2013 - Copyright Pierre-Yves Chibon

 Distributed under License GPLv3 or later
 You can find a copy of this license on the website
 http://www.gnu.org/licenses/gpl.html

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 MA 02110-1301, USA.

 MQ² test script for the Excel plugin
"""

import os
import shutil
import sys
import unittest

from datetime import date

sys.path.insert(0, os.path.abspath('..'))

import MQ2
import MQ2.mq2 as mq2


TEST_INPUT_PASSED = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'xls', 'rqtl_out.xls.zip')
TEST_INPUT_FAILED = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'xls', 'rqtl_out.xls.invalid.zip')
TEST_INPUT_FAILED2 = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'xls', 'rqtl_out.xls.invalid2.zip')
TEST_INPUT_FAKE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'invalid', 'fake.zip')

TEST_INPUT_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'xls', 'rqtl_out.xls')

TEST_FAKE_INPUT_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'xls', 'fake.xls')

TEST_FOLDER = os.path.dirname(os.path.abspath(__file__))


def clean_directory():
    """ Remove the 'demo_test' folder is present in the current
    working directory.
    """
    if os.path.exists('demo_test') and os.path.isdir('demo_test'):
        shutil.rmtree('demo_test')
    for filename in os.listdir('.'):
        if filename.endswith(('.map', '.xls', '.xlsx', '.csv')):
            os.unlink(filename)


def read_file(filename):
    """ Reads in a file for a given filename and just return the
    content.
    """
    stream = open(filename)
    data = stream.read()
    stream.close()
    return data


class MQ2MapQTLtests(unittest.TestCase):
    """ MQ² tests for MapQTL input. """

    def __init__(self, method_name='runTest'):
        """ Constructor. """
        unittest.TestCase.__init__(self, method_name)

    def setUp(self):
        """ Set up the environnment, ran before every tests. """
        clean_directory()

    def tearDown(self):
        """ Clean up the environnment, ran after every tests. """
        clean_directory()

    def test_get_plugin_and_folder(self):
        """ Test the get_plugin_and_folder function with MapQTL zip
        input.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertTrue(os.path.basename(
            folder).startswith(date.today().strftime('%Y%m%d')))
        self.assertEqual(plugin.name, 'Excel plugin')
        self.assertEqual(plugin.session_name, 'Excel sheet')

    def test_get_plugin_and_folder_wrong_input(self):
        """ Test the get_plugin_and_folder function with an invalid
        zip input.
        """
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.get_plugin_and_folder,
            inputzip=TEST_INPUT_FAKE)

    def test_run_mq2_no_session(self):
        """ Test the run_mq2 function with MapQTL zip input and without
        session identifier.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertRaises(
            MQ2.MQ2NoSessionException,
            mq2.run_mq2,
            plugin, folder, lod_threshold=3, session=None)

    def test_run_mq2_no_folder(self):
        """ Test the run_mq2 function with MapQTL zip input and without
        folder.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.run_mq2,
            plugin, folder=None, lod_threshold=3, session=2)

    def test_run_mq2_invalid_lod(self):
        """ Test the run_mq2 function with MapQTL zip input and an
        invalid LOD threshold.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertRaises(
            MQ2.MQ2Exception,
            mq2.run_mq2,
            plugin, folder=folder, lod_threshold='a', session='Sheet1')

    def test_run_mq2_invalid_session(self):
        """ Test the run_mq2 function with MapQTL zip input and the
        wrong session identifier.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertRaises(
            MQ2.MQ2NoSuchSessionException,
            mq2.run_mq2,
            plugin, folder, lod_threshold=3, session='Sheet2')

    def test_run_mq2(self):
        """ Test the run_mq2 function from a zip archive.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        mq2.run_mq2(plugin, folder, lod_threshold=3, session='Sheet1')
        self.assertEqual(read_file('qtls.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'qtls.exp')))
        self.assertEqual(read_file('map.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'map.exp')))
        self.assertEqual(read_file('map_with_qtls.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'map_with_qtls.exp')))
        self.assertEqual(read_file('qtls_with_mk.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'qtls_with_mk.exp')))
        ## Checks disabled
        ## Due to python3 or xlrd on python3 compatibility issue
        ## the number of significant number after the commas differs
        ## which of course breaks the check
        #self.assertEqual(read_file('qtls_matrix.csv'),
                         #read_file(os.path.join(
                            #TEST_FOLDER, 'xls', 'qtls_matrix.exp')))
        #self.assertEqual(read_file('MapChart.map'),
                         #read_file(os.path.join(
                            #TEST_FOLDER, 'xls', 'MapChart.exp')))

    def test_run_mq2_from_file(self):
        """ Test the run_mq2 function from a file.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputfile=TEST_INPUT_FILE)
        mq2.run_mq2(plugin,
                    folder=TEST_INPUT_FILE,
                    lod_threshold=3,
                    session='Sheet1')
        self.assertEqual(read_file('qtls.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'qtls.exp')))
        self.assertEqual(read_file('map.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'map.exp')))
        self.assertEqual(read_file('map_with_qtls.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'map_with_qtls.exp')))
        self.assertEqual(read_file('qtls_with_mk.csv'),
                         read_file(os.path.join(
                            TEST_FOLDER, 'xls', 'qtls_with_mk.exp')))
        ## Checks disabled
        ## Due to python3 or xlrd on python3 compatibility issue
        ## the number of significant number after the commas differs
        ## which of course breaks the check
        #self.assertEqual(read_file('qtls_matrix.csv'),
                         #read_file(os.path.join(
                            #TEST_FOLDER, 'xls', 'qtls_matrix.exp')))
        #self.assertEqual(read_file('MapChart.map'),
                         #read_file(os.path.join(
                            #TEST_FOLDER, 'xls', 'MapChart.exp')))

    def test_plugin_valid_file(self):
        """ Test the valid_file method of the plugin.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertFalse(plugin.valid_file(TEST_INPUT_PASSED))
        self.assertTrue(plugin.valid_file(
            os.path.join(folder,os.listdir(folder)[0])))

        self.assertFalse(plugin.valid_file(
            TEST_FAKE_INPUT_FILE))

    def test_plugin_get_files(self):
        """ Test the get_files method of the plugin.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertEqual(len(plugin.get_files(folder)), 1)
        self.assertTrue(
            plugin.get_files(folder)[0].endswith('/rqtl_out.xls'))
        self.assertEqual(plugin.get_files(None), [])

    def test_plugin_get_session_identifiers_failed(self):
        """ Test the get_session_identifiers method of the plugin.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertRaises(
            MQ2.MQ2Exception,
            plugin.get_session_identifiers,
            folder=os.path.join(TEST_FOLDER, 'fake.xls'),
            inputfile=folder)

    def test_plugin_get_session_identifiers(self):
        """ Test the get_session_identifiers method of the plugin.
        """
        plugin, folder = mq2.get_plugin_and_folder(
            inputzip=TEST_INPUT_PASSED)
        self.assertEqual(len(plugin.get_session_identifiers(folder)), 1)
        self.assertEqual(plugin.get_session_identifiers(folder), ['Sheet1'])
        self.assertEqual(plugin.get_session_identifiers(None), [])

        self.assertEqual(plugin.get_session_identifiers(
            folder=os.path.join(TEST_FOLDER, 'fake.xls')), [])

        self.assertEqual(plugin.get_session_identifiers(
            inputfile=folder), [])


if __name__ == '__main__':
    SUITE = unittest.TestLoader().loadTestsFromTestCase(MQ2MapQTLtests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)
